
//
// Copyright (c) 2014, Peter A.M. van Hoof.
//
//    This program is free software: you can redistribute it and/or modify
//    it under the terms of the GNU General Public License as published by
//    the Free Software Foundation, either version 3 of the License, or
//    (at your option) any later version.
//
//    This program is distributed in the hope that it will be useful,
//    but WITHOUT ANY WARRANTY; without even the implied warranty of
//    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//    GNU General Public License for more details.
//
//    You should have received a copy of the GNU General Public License
//    along with this program.  If not, see <http://www.gnu.org/licenses/>.
//
// Peter A.M. van Hoof
// Royal Observatory of Belgium
// Ringlaan 3
// B-1180 Brussels
// Belgium
// p.vanhoof@oma.be
//
//
// If you use any of these data in a scientific publication, please refer to
//
// van Hoof P.A.M., Williams R.J.R., Volk K., Chatzikos M., Ferland G.J., Lykins M., Porter R.L., Wang Y.
// Accurate determination of the free-free Gaunt factor, I -- non-relativistic Gaunt factors
// 2014, MNRAS, 444, 420
//

#include "mpreal.h"
#include <iostream>
#include <iomanip>
#include <cstdlib>
#include <complex>
#include <vector>
#include <algorithm>
#include <mpi.h>
#include <cassert>
#include <fstream>
#include <map>
#include <unistd.h>
#include <sys/file.h>

// uncomment the following line to compute a table of non-averaged Gaunt factors
// #define GENERATE_NONAV 1

// uncomment the following line to compute a table of thermally averaged Gaunt factors
#define GENERATE_THERM_AV 1

// uncomment the following line to compute a table of total (frequency integrated) Gaunt factors
// #define GENERATE_TOTAL 1

using namespace std;
using namespace mpfr;

static const mp_prec_t prec_start = 128;
static const mp_prec_t prec_max = 4096;
static const double max_rerr = 1.e-15;

template<typename T>
T pow2(T x) { return x*x; }

inline bool finite(const mpreal& x) { return _isnum(x); }

inline mpreal powi(const mpreal& x, int i)
{
	assert( i >= 0 );

	mpreal res( 1 );
	mpreal powx = x;
	while( true ) {
		if( i%2 == 1 )
			res *= powx;
		i >>= 1;
		if( i == 0 )
			break;
		powx *= powx;
	}
	return res;
}

inline mpreal operator "" _mp( unsigned long long l )
{
	return mpreal(size_t(l));
}
inline mpreal operator "" _mp( long double l )
{
	return mpreal(l);
}
inline mpreal operator "" _mp( const char* s, size_t )
{
	return mpreal(s);
}

class loss_of_precision
{
public:
	loss_of_precision() {}
};

typedef complex<mpreal> mpcomplex;

double g_iii_fr_int(const mpreal&,double&,const double&);
mpreal g_iii_av(const mpreal&,const mpreal&,mpreal&,const mpreal&);
void gaunt_step(const mpreal&,const mpreal&,mpreal(*)(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&),
		mpreal&,mpreal&,const mpreal&,mpreal&,mpreal&,int&);
void gaunt_step2(const mpreal&,const mpreal&,mpreal(*)(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&),
		 double&,double&,const double&,double&,double&,int&);
mpreal g_fun(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&);
mpreal g_fun2(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&);
mpreal g_iii(const mpreal&,const mpreal&,mpreal&);
void Coulomb(const mpreal&,const mpreal&,const mpreal&,mpreal&,mpreal&,mpreal&,mpreal&);
mpreal Il(int,const mpreal&,const mpreal&,const mpreal&,mpreal&,const mpcomplex&,const mpcomplex&,const mpcomplex&);
mpreal Gl(int,const mpreal&,const mpreal&,const mpreal&,mpreal&,const mpcomplex&,const mpcomplex&,const mpcomplex&,
	  mpreal&);
mpcomplex F21(const mpcomplex&,const mpcomplex&,const mpcomplex&,const mpreal&,mpreal&);
mpcomplex clgamma(const mpcomplex&);
mpreal eta_sub(const mpreal&,const mpreal&);

unsigned int nCPU;
unsigned int nRANK;
unsigned int nGROUP;
unsigned int ngrRANK;

MPI_Comm GroupComm;

int main(int argc, char* argv[])
{
	mpreal::set_default_prec(prec_start);

#if defined(GENERATE_NONAV)
	MPI_Init( &argc, &argv );
	int arg;
	MPI_Comm_size( MPI_COMM_WORLD, &arg );
	nCPU = arg;
	MPI_Comm_rank( MPI_COMM_WORLD, &arg );
	nRANK = arg;

	unsigned int g2max = 151;
	unsigned int umax = 276;

	mpreal g2start = -20_mp;
	mpreal ustart = -30_mp;
	mpreal steps_per_dex = 5_mp;

	vector<double> resv( g2max*umax );
	vector<double> dresv( g2max*umax );
	for( unsigned int i=0; i < resv.size(); ++i ) {
		resv[i] = 0.;
		dresv[i] = 0.;
	}

	mp_prec_t myprec;
	if( nRANK == 0 ) {
		// this rank does not do any work, it only sends out jobs to the other ranks
		vector<int> jobs( g2max*umax + nCPU-1 );
		unsigned int k = 0;
		for( unsigned int g2=0; g2 < g2max; ++g2 )
			for( unsigned int u=0; u < umax; ++u )
				jobs[k++] = u*g2max + g2;
		// this is the end-of-jobs signal for the other ranks
		for( unsigned int i=1; i < nCPU; ++i )
			jobs[k++] = -1;

		assert( k == jobs.size() );

		for( unsigned int k=0; k < jobs.size(); ++k ) {
			int dest = -1;
			MPI_Status status;
			MPI_Recv( &dest, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
			int job = jobs[k];
			MPI_Send( &job, 1, MPI_INT, dest, 0, MPI_COMM_WORLD );
		}
	}
	else {
		// these ranks do the actual work
		while( true ) {
			MPI_Status status;
			MPI_Send( &nRANK, 1, MPI_UNSIGNED, 0, 0, MPI_COMM_WORLD );
			int job;
			MPI_Recv( &job, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
			if( job < 0 )
				break;
			int g2 = job%g2max;
			int u = job/g2max;
			mpreal gam2 = mpreal(g2)/steps_per_dex + g2start;
			mpreal uu = mpreal(u)/steps_per_dex + ustart;
			mpreal eps_i = exp10(gam2);
			mpreal w = exp10(uu);
			mpreal err, res;
			res = exp(1_mp)*g_fun( 1_mp, 1_mp/eps_i, w/eps_i, err, myprec );
			resv[ job ] = double(res);
			dresv[ job ] = double(max(err,2.2250738585072014e-308_mp));
		}
	}

	MPI_Barrier( MPI_COMM_WORLD );

	vector<double> resc = resv;
	MPI_Reduce( &resc[0], &resv[0], resc.size(), MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
	vector<double> dresc = dresv;
	MPI_Reduce( &dresc[0], &dresv[0], resc.size(), MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

	MPI_Barrier( MPI_COMM_WORLD );

	if( nRANK == 0 ) {
		cout << setw(9) << 20140210 << " # magic number" << endl;
		cout << setw(4) << g2max << " " << setw(4) << umax;
		cout << " # number of grid points in eps_i and w direction" << endl;
		cout << setw(9) << g2start << " # start value for log(eps_i)" << endl;
		cout << setw(9) << ustart << " # start value for log(w)" << endl;
		cout << setw(9) << 1_mp/steps_per_dex << " # stepsize in dex for log(eps_i) and log(w)" << endl;
		cout << "#" << endl;
		cout << "# Rows have constant log(w), columns have constant log(eps_i)" << endl;
		for( unsigned int k=0; k < resv.size(); ++k ) {
			cout << scientific << setprecision(10) << setw(18) << resv[k];
			if( (k+1)%g2max == 0 ) {
				cout << endl;
			}
		}
	}

	MPI_Finalize();

	return 0;
#elif defined(GENERATE_THERM_AV)
	MPI_Init( &argc, &argv );
	int arg;
	MPI_Comm_size( MPI_COMM_WORLD, &arg );
	nCPU = arg;
	MPI_Comm_rank( MPI_COMM_WORLD, &arg );
	nRANK = arg;

	unsigned int g2max = 81;
	unsigned int umax = 146;

	mpreal g2start = -6_mp;
	mpreal ustart = -16_mp;
	mpreal steps_per_dex = 5_mp;

	vector<double> resv( g2max*umax );
	vector<double> dresv( g2max*umax );
	for( unsigned int i=0; i < resv.size(); ++i ) {
		resv[i] = 0.;
		dresv[i] = 0.;
	}

	if( nRANK == 0 ) {
		// this rank does not do any work, it only sends out jobs to the other ranks
		vector<int> jobs( g2max*umax + nCPU-1 );
		unsigned int k = 0;
		// jobs for high gamma^2 tend to be longer running, so we do those first
		for( int g2=g2max-1; g2 >= 0; --g2 )
			for( unsigned int u=0; u < umax; ++u )
				jobs[k++] = u*g2max + g2;
		// this is the end-of-jobs signal for the other ranks
		for( unsigned int i=1; i < nCPU; ++i )
			jobs[k++] = -1;

		assert( k == jobs.size() );

		for( unsigned int k=0; k < jobs.size(); ++k ) {
			int dest = -1;
			MPI_Status status;
			MPI_Recv( &dest, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
			int job = jobs[k];
			MPI_Send( &job, 1, MPI_INT, dest, 0, MPI_COMM_WORLD );
		}
	}
	else {
		// these ranks do the actual work
		while( true ) {
			MPI_Status status;
			MPI_Send( &nRANK, 1, MPI_UNSIGNED, 0, 0, MPI_COMM_WORLD );
			int job;
			MPI_Recv( &job, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
			if( job < 0 )
				break;
			int g2 = job%g2max;
			int u = job/g2max;
			mpreal gam2 = mpreal(g2)/steps_per_dex + g2start;
			mpreal uu = mpreal(u)/steps_per_dex + ustart;
			mpreal err, res;
			res = g_iii_av( exp10(gam2), exp10(uu), err, 1.e-5_mp );
			resv[ job ] = double(res);
			dresv[ job ] = double(err);
		}
	}

	MPI_Barrier( MPI_COMM_WORLD );

	vector<double> resc = resv;
	MPI_Reduce( &resc[0], &resv[0], resc.size(), MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );
	vector<double> dresc = dresv;
	MPI_Reduce( &dresc[0], &dresv[0], resc.size(), MPI_DOUBLE, MPI_SUM, 0, MPI_COMM_WORLD );

	MPI_Barrier( MPI_COMM_WORLD );

	if( nRANK == 0 ) {
		cout << setw(9) << 20140210 << " # magic number" << endl;
		cout << setw(4) << g2max << " " << setw(4) << umax;
		cout << " # number of grid points in gam2 and u direction" << endl;
		cout << setw(9) << g2start << " # start value for log(gam2)" << endl;
		cout << setw(9) << ustart << " # start value for log(u)" << endl;
		cout << setw(9) << 1_mp/steps_per_dex << " # stepsize in dex for log(gam2) and log(u)" << endl;
		cout << "#" << endl;
		cout << "# thermally averaged Gaunt factors. Rows have constant log(u),";
		cout << " columns have constant log(gam2)" << endl;
		for( unsigned int k=0; k < resv.size(); ++k ) {
			cout << scientific << setprecision(8) << setw(16) << resv[k];
			if( (k+1)%g2max == 0 )
				cout << endl;
		}
		cout << "#" << endl;
		cout << "# uncertainties in the thermally averaged Gaunt factors" << endl;
		for( unsigned int k=0; k < resv.size(); ++k ) {
			cout << scientific << setprecision(4) << setw(12) << dresv[k];
			if( (k+1)%g2max == 0 )
				cout << endl;
		}
	}

	MPI_Finalize();

	return 0;
#elif defined(GENERATE_TOTAL)
	MPI_Init( &argc, &argv );
	int arg;
	MPI_Comm_size( MPI_COMM_WORLD, &arg );
	nCPU = arg;
	MPI_Comm_rank( MPI_COMM_WORLD, &arg );
	nRANK = arg;

	assert( 4*(nCPU/4) == nCPU );

	nGROUP = nRANK/4;
	MPI_Comm_split(MPI_COMM_WORLD, nGROUP, nRANK, &GroupComm);

	MPI_Comm_rank( GroupComm, &arg );
	ngrRANK = arg;

	unsigned int g2max = 81;

	mpreal g2start = -6_mp;
	mpreal steps_per_dex = 5_mp;

	vector<double> resv( g2max );
	vector<double> dresv( g2max );
	for( unsigned int i=0; i < resv.size(); ++i ) {
		resv[i] = 0.;
		dresv[i] = 0.;
	}

	if( nRANK == 0 ) {
		// this rank does not do any work, it only sends out jobs to the other ranks
		vector<int> jobs( g2max + nCPU/4 - 1 );
		unsigned int k = 0;
		// jobs for high gamma^2 tend to be longer running, so we do those first
		for( int g2=g2max-1; g2 >= 0; --g2 )
			jobs[k++] = g2;
		// this is the end-of-jobs signal for the other ranks
		for( unsigned int i=1; i < nCPU/4; ++i )
			jobs[k++] = -1;

		assert( k == jobs.size() );

		for( unsigned int k=0; k < jobs.size(); ++k ) {
			int dest = -1;
			MPI_Status status;
			MPI_Recv( &dest, 1, MPI_INT, MPI_ANY_SOURCE, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
			int job = jobs[k];
			MPI_Send( &job, 1, MPI_INT, dest, 0, MPI_COMM_WORLD );
		}
	}
	else {
		// these ranks do the actual work
		while( nRANK >= 4 ) {
			int job;
			if( ngrRANK == 0 ) {
				MPI_Status status;
				MPI_Send( &nRANK, 1, MPI_UNSIGNED, 0, 0, MPI_COMM_WORLD );
				MPI_Recv( &job, 1, MPI_INT, 0, MPI_ANY_TAG, MPI_COMM_WORLD, &status );
			}
			MPI_Bcast( &job, 1, MPI_INT, 0, GroupComm );
			if( job < 0 )
				break;
			int g2 = job;
			mpreal gam2 = mpreal(g2)/steps_per_dex + g2start;
			double err, res;
			res = g_iii_fr_int( exp10(gam2), err, 1.e-5 );
			if( ngrRANK == 0 ) {
				resv[ job ] = res;
				dresv[ job ] = err;
			}
		}
	}

	MPI_Barrier( MPI_COMM_WORLD );

	vector<double> resc = resv;
	MPI_Reduce( &resc[0], &resv[0], resc.size(), MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );
	vector<double> dresc = dresv;
	MPI_Reduce( &dresc[0], &dresv[0], resc.size(), MPI_DOUBLE, MPI_MAX, 0, MPI_COMM_WORLD );

	MPI_Barrier( MPI_COMM_WORLD );

	if( nRANK == 0 ) {
		cout << setw(9) << 20140424 << " # magic number" << endl;
		cout << setw(9) << g2max << " # number of grid points in gam2 direction" << endl;
		cout << setw(9) << g2start << " # start value for log(gam2)" << endl;
		cout << setw(9) << 1_mp/steps_per_dex << " # stepsize in dex for log(gam2)" << endl;
		cout << "#" << endl;
		cout << "# log(gam^2)  <g_ff(gam^2)>  unc" << endl;
		for( unsigned int k=0; k < resv.size(); ++k ) {
			cout << fixed << setprecision(2) << setw(6) << mpreal(k)/steps_per_dex + g2start;
			cout << scientific << setprecision(8) << setw(16) << resv[k];
			cout << scientific << setprecision(4) << setw(12) << dresv[k] << endl;
		}
	}

	MPI_Finalize();

	return 0;
#else
#error "uncomment one of the GENERATE_XXX statements near the top of the file gauntff_nr.cc"
#endif
}

double g_iii_fr_int(const mpreal& gam2, double& err, const double& tol)
{
	mpreal u = 0_mp;
	double x = 0.;
	double h = max(sqrt(double(gam2)/15.), 1.)*min(1.e-5, tol);
	double res = 1.;
	err = 0.;
	// error == 0 : integration needs more steps
	// error == 1 : normal termination of integral
	// error == 2 : step failed to converge
	int error = 0;
	do {
		gaunt_step2( gam2, u, g_fun2, x, h, tol/2., res, err, error );
	} while( error == 0 );

	if( error < 2 )
		return res - 1.;
	else {
		cerr << "integration failed: gam2 = " << gam2 << " u = " << u << endl;
		throw loss_of_precision();
	}
}

mpreal g_iii_av(const mpreal& gam2, const mpreal& u, mpreal& err, const mpreal& tol)
{
	mpreal x = 0_mp;
	mpreal h = min(sqrt(2.e4_mp*u), 1_mp)*max(sqrt(gam2/15_mp), 1_mp)*min(1.e-5_mp, tol);
	mpreal res = 0_mp;
	err = 0_mp;
	// error == 0 : integration needs more steps
	// error == 1 : normal termination of integral
	// error == 2 : step failed to converge
	int error = 0;
	do {
		gaunt_step( gam2, u, g_fun, x, h, tol, res, err, error );
	} while( error == 0 );

	if( error < 2 )
		return res;
	else {
		cerr << "integration failed: gam2 = " << gam2 << " u = " << u << endl;
		throw loss_of_precision();
	}
}

void gaunt_step(const mpreal& gam2,
		const mpreal& u,
		mpreal (*fun)(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&),
		mpreal& x,
		mpreal& h,
		const mpreal& toler,
		mpreal& yval,
		mpreal& dyval,
		int& error)
{
	if( error > 0 )
		return;

	const int MAX_ITER = 8;
	const int MAX_DIVS = 4;
	const int MAX_ABS = (1<<MAX_DIVS) + 1;

	mpreal y1[MAX_DIVS+1];
	mpreal dy1[MAX_DIVS+1];
	mpreal y1a[MAX_ABS];
	mpreal y1b[MAX_ABS];

	for( int k=0; k < MAX_ITER; ++k )
	{
		for( int i=0; i <= MAX_DIVS; ++i )
		{
			int nabs = 1<<i;
			int stride = 1<<(MAX_DIVS-i);
			mpreal h1 = h/mpreal(nabs);
			// for k > 0 the points at 0, MAX_DIVS/2 and MAX_DIVS are already set up...
			if( k == 0 || i > 1 )
			{
				for( int j=1; j <= nabs; ++j )
				{
					int index = j*stride;
					// y1a[index] may already have been initialized on a previous
					// iteration of the loop over i. If so, do not repeat the calculation
					if( i == 0 || j%2 == 1 )
					{
						mpreal x1 = x + h1*mpreal(j);
						mpreal rerr;
						mp_prec_t myprec;
						y1a[index] = (*fun)(x1,gam2,u,rerr,myprec);
						y1b[index] = rerr*y1a[index];
					}
				}
			}
			if( i < 2 ) {
				y1[i] = 0_mp;
				dy1[i] = 0_mp;
			}
			else {
				// use Eq. 4.1.20 from Numerical Recipes in C, 1st ed.
				y1[i] = (23_mp*y1a[stride]+7_mp*y1a[2*stride]+
					 13_mp*y1a[(nabs-1)*stride]+5_mp*y1a[nabs*stride])/12_mp;
				for( int j=3; j < nabs-1; ++j )
					y1[i] += y1a[j*stride];
				y1[i] *= h1;
				// this is the regular error propagation formula assuming errors are uncorrelated
				dy1[i] = (pow2(23_mp*y1b[stride])+pow2(7_mp*y1b[2*stride])+
					  pow2(13_mp*y1b[(nabs-1)*stride])+pow2(5_mp*y1b[nabs*stride]))/144_mp;
				for( int j=3; j < nabs-1; ++j )
					dy1[i] += pow2(y1b[j*stride]);
				dy1[i] = sqrt(dy1[i])*h1;
			}
			// check for convergence by comparing the integrals with stepsizes h1 and 2*h1
			if( i > 2 )
			{
				// toler is relative precision
				mpreal err = abs((y1[i]-y1[i-1])/((yval+y1[i])*toler));
				if( err <= 1_mp )
				{
					mpreal err = hypot(dy1[i],abs(y1[i]-y1[i-1])/2_mp);
					yval += y1[i];
					dyval = hypot(dyval,err);
					mpreal fac = 0.9_mp*cbrt(1_mp/max(err,0.1_mp));
					x += h;
					h = 8_mp*h1*fac;
					// make estimate of the remainder of the integral to infinity
					// assuming that g_iii() is constant
					// use additional factor 100 to make sure that the error from
					// truncating the integral is negligible compared to the other
					// terms and our error estimate is valid
					error = ( abs(y1[i]/(yval*expm1(h))) < toler/100_mp ) ? 1 : 0;
					return;
				}
			}
		}
		// convergence failed. Here we reduce the stepsize and try again...
		y1a[(MAX_ABS-1)/2] = y1a[1];
		y1a[MAX_ABS-1] = y1a[2];
		y1b[(MAX_ABS-1)/2] = y1b[1];
		y1b[MAX_ABS-1] = y1b[2];
		h /= mpreal((MAX_ABS-1)/2);
	}
	error = 2;
	return;
}

void gaunt_step2(const mpreal& gam2,
		 const mpreal& u,
		 mpreal (*fun)(const mpreal&,const mpreal&,const mpreal&,mpreal&,mp_prec_t&),
		 double& x,
		 double& h,
		 const double& toler,
		 double& yval,
		 double& dyval,
		 int& error)
{
	if( error > 0 )
		return;

	const int MAX_ITER = 8;
	const int MAX_DIVS = 4;
	const int MAX_ABS = (1<<MAX_DIVS) + 1;

	vector<double> y1(MAX_DIVS+1);
	vector<double> dy1(MAX_DIVS+1);
	vector<double> y1a(MAX_ABS);
	vector<double> y1b(MAX_ABS);

	for( int k=0; k < MAX_ITER; ++k )
	{
		unsigned int nn = 0;
		for( int i=0; i <= MAX_DIVS; ++i )
		{
			int nabs = 1<<i;
			int stride = 1<<(MAX_DIVS-i);
			mpreal h1 = mpreal(h)/mpreal(nabs);
			// for k > 0 the points at 0, MAX_DIVS/2 and MAX_DIVS are already set up...
			if( k == 0 || i > 1 )
			{
				for( int j=1; j <= nabs; ++j )
				{
					int index = j*stride;
					// y1a[index] may already have been initialized on a previous
					// iteration of the loop over i. If so, do not repeat the calculation
					if( i == 0 || j%2 == 1 )
					{
						mpreal x1 = x + h1*mpreal(j);
						mpreal rerr;
						mp_prec_t myprec;
						if( ((nn++)%4) == ngrRANK ) {
							y1a[index] = double((*fun)(x1,gam2,u,rerr,myprec));
							y1b[index] = double(rerr*y1a[index]);
						}
						else {
							y1a[index] = 0.;
							y1b[index] = 0.;
						}
					}
				}
			}
			if( i < 2 ) {
				y1[i] = 0.;
				dy1[i] = 0.;
			}
			else {
				MPI_Barrier( GroupComm );

				vector<double> cy1a = y1a;
				MPI_Reduce( &cy1a[0], &y1a[0], y1a.size(), MPI_DOUBLE, MPI_MAX, 0, GroupComm );
				MPI_Bcast( &y1a[0], y1a.size(), MPI_DOUBLE, 0, GroupComm );
				vector<double> cy1b = y1b;
				MPI_Reduce( &cy1b[0], &y1b[0], y1b.size(), MPI_DOUBLE, MPI_MAX, 0, GroupComm );
				MPI_Bcast( &y1b[0], y1b.size(), MPI_DOUBLE, 0, GroupComm );
				nn = 0;

				MPI_Barrier( GroupComm );

				// use Eq. 4.1.20 from Numerical Recipes in C, 1st ed.
				y1[i] = (23.*y1a[stride]+7.*y1a[2*stride]+
					 13.*y1a[(nabs-1)*stride]+5.*y1a[nabs*stride])/12.;
				for( int j=3; j < nabs-1; ++j )
					y1[i] += y1a[j*stride];
				y1[i] *= double(h1);
				// this is the regular error propagation formula assuming errors are uncorrelated
				dy1[i] = (pow2(23.*y1b[stride])+pow2(7.*y1b[2*stride])+
					  pow2(13.*y1b[(nabs-1)*stride])+pow2(5.*y1b[nabs*stride]))/144.;
				for( int j=3; j < nabs-1; ++j )
					dy1[i] += pow2(y1b[j*stride]);
				dy1[i] = sqrt(dy1[i])*double(h1);
			}
			// check for convergence by comparing the integrals with stepsizes h1 and 2*h1
			if( i > 2 )
			{
				// toler is relative precision
				double err = abs((y1[i]-y1[i-1])/((yval+y1[i])*toler));
				if( err <= 1. )
				{
					double err = hypot(dy1[i],abs(y1[i]-y1[i-1])/2.);
					yval += y1[i];
					dyval = hypot(dyval,err);
					double fac = 0.9*cbrt(1./max(err,0.1));
					x += h;
					h = 8.*h1*fac;
					// make estimate of the remainder of the integral to infinity
					// assuming that g_iii() is constant
					// use additional factor 100 to make sure that the error from
					// truncating the integral is negligible compared to the other
					// terms and our error estimate is valid
					error = ( abs(y1[i]/(yval*expm1(h))) < toler/100. ) ? 1 : 0;
					return;
				}
			}
		}
		// convergence failed. Here we reduce the stepsize and try again...
		y1a[(MAX_ABS-1)/2] = y1a[1];
		y1a[MAX_ABS-1] = y1a[2];
		y1b[(MAX_ABS-1)/2] = y1b[1];
		y1b[MAX_ABS-1] = y1b[2];
		h /= double((MAX_ABS-1)/2);
	}
	error = 2;
	return;
}

// g_fun: helper function for calculating the temperature-averaged gaunt factor
mpreal g_fun(const mpreal& x, const mpreal& gam2, const mpreal& u, mpreal& rerr, mp_prec_t& myprec)
{
	bool lgConverged = false;
	mpreal eps_i = x/gam2;
	mpreal eta_i = 1_mp/sqrt(eps_i);
	mpreal w = u/gam2;
	mpreal kol2 = eps_i/(eps_i+w);
	mpreal k = 1_mp/sqrt(eps_i+w);
	mpreal res;

	mpreal err = 0.025_mp/cbrt(powi((1_mp - kol2)*k,8));
	if( err > 1.e-14_mp || w > 1.e-6_mp || sqrt(powi(eps_i,3))/w > 1.e-4_mp ) {
		mp_prec_t save = mpreal::default_prec;
		myprec = save;

		do {
			mpreal::set_default_prec(myprec);

			// make local copies at higher precision
			// keep declaration and initialization separate in order
			// to assure that the variables indeed get double precision...
			mpreal eps_i, w;
			eps_i = x/gam2;
			w = u/gam2;

			try {
				res = g_iii( eps_i, w, rerr );
				lgConverged = ( rerr <= max_rerr );
			}
			catch( loss_of_precision& ) {
				lgConverged = false;
			}

			myprec *= 2;
		} while( !lgConverged && myprec <= prec_max );

		myprec /= 2; // undo the last multiplication by 2
		mpreal::set_default_prec(save);
		rerr = max(rerr,machine_epsilon(save));
	}

	if( !lgConverged ) {
		// this is a fixed version of equation 1.41 of Menzel D.H. & Pekeris C.L., 1935, MNRAS, 96, 77
		// also an additional higher-order term has been added
		res = 1_mp + 0.17282603693599267453_mp*(1_mp + kol2)/cbrt(pow2((1_mp - kol2)*k)) -
			0.049595701685869587165_mp*(1_mp - 4_mp/3_mp*kol2 + pow2(kol2))/cbrt(powi((1_mp - kol2)*k,4)) -
			3_mp/175_mp*(1_mp - kol2/3_mp - pow2(kol2)/3_mp + powi(kol2,3))/pow2((1_mp - kol2)*k);
		rerr = err/res;
		myprec = 1;
		// this is the region where the series expansion above is valid
		assert( w <= 1.e-6_mp && sqrt(powi(eps_i,3))/w <= 1.e-4_mp );
	}

	return exp(-x)*res;
}

mpreal g_fun2(const mpreal& x, const mpreal& gam2, const mpreal&, mpreal& rerr, mp_prec_t& myprec)
{
	mpreal err;
	mpreal res = g_iii_av(gam2,x,err,1.e-6_mp);
	rerr = err/res;
	myprec = 0;
	return exp(-x)*res;
}

// g_iii: this implements Eq. 16 of Karzas & Latter, 1961, ApJS 6, 167
mpreal g_iii(const mpreal& eps_i, const mpreal& w, mpreal& rerr)
{
	mpreal rerr1, rerr2;
	mpreal pi = 2_mp*asin(1_mp);

	mpreal eta_i = 1_mp/sqrt(eps_i);
	mpreal eta_f = 1_mp/sqrt(eps_i + w);
	mpreal eta_imf = eta_sub( eps_i, w );
	mpreal r0, r1;
	Coulomb( eta_i, eta_f, eta_imf, r0, r1, rerr1, rerr2 );
	mpreal c1 = pow2(eta_i) + pow2(eta_f) + 2_mp*pow2(eta_i)*pow2(eta_f);
	mpreal c2 = 2_mp*eta_i*eta_f*sqrt(1_mp + pow2(eta_i))*sqrt(1_mp + pow2(eta_f));
	mpreal res = 2_mp*sqrt(3_mp)/(pi*eta_i*eta_f)*(c1*r0 - c2*r1)*r0;
	rerr2 = hypot(c1*r0*rerr1,c2*r1*rerr2)/abs(c1*r0 - c2*r1);
	rerr = hypot( rerr1, rerr2 );
	if( !finite(res) || rerr > max_rerr )
		throw loss_of_precision();
	return res;
}

void Coulomb(const mpreal& eta_i, const mpreal& eta_f, const mpreal& eta_imf,
	     mpreal& r0, mpreal& r1, mpreal& rerr0, mpreal& rerr1)
{
	mpreal x = 4_mp*eta_i*eta_f/pow2(eta_imf);

	mpcomplex g1( 1, eta_i );
	mpcomplex g2( 1, eta_f );
	mpcomplex cgi0 = clgamma( g1 );
	// use recursion for log(Gamma( (2, eta_i) ))
	mpcomplex cgi1 = cgi0 + log(g1);
	mpcomplex cgf0 = clgamma( g2 );
	// use recursion for log(Gamma( (2, eta_f) ))
	mpcomplex cgf1 = cgf0 + log(g2);
	mpcomplex cg2( 0, 0 );
	if( abs(x) > 1_mp ) {
		mpcomplex g3( 0, -eta_imf );
		cg2 = clgamma( g3 );
	}
	r0 = Il( 0, eta_i, eta_f, eta_imf, rerr0, cgi0, cgf0, cg2 );
	r1 = Il( 1, eta_i, eta_f, eta_imf, rerr1, cgi1, cgf1, cg2 );
}

// Il: this implements Eq. 9 of Karzas & Latter, 1961, ApJS 6, 167
mpreal Il(int l, const mpreal& eta_i, const mpreal& eta_f, const mpreal& eta_imf, mpreal& rerr,
	  const mpcomplex& cgi, const mpcomplex& cgf, const mpcomplex& cg2)
{
	assert( l == 0 || l == 1 );

	mpreal x = 4_mp*eta_i*eta_f/pow2(eta_imf);
	mpreal pi = 2_mp*asin(1_mp);

	mpreal cg;
	mpreal e1( pi*abs(eta_imf)/2_mp );
	mpreal res = Gl( l, eta_i, eta_f, eta_imf, rerr, cgi, cgf, cg2, cg );
	res *= exp((cgi + cgf + cg + e1).real());
	// divide by Gamma(2*l+2)
	if( l == 1 )
		res /= 6_mp;
	res *= powi(x,l+1)/4_mp;
	return res;
}

// Gl: this implements Eqs. 10 & 11 of Karzas & Latter, 1961, ApJS 6, 167
mpreal Gl(int l, const mpreal& eta_i, const mpreal& eta_f, const mpreal& eta_imf, mpreal& rerr,
	  const mpcomplex& cgi, const mpcomplex& cgf, const mpcomplex& cg2,
	  mpreal& cg)
{
	assert( l == 0 || l == 1 );

	mpreal x = -4_mp*eta_i*eta_f/pow2(eta_imf);
	mpreal xp = abs((eta_imf)/(eta_i+eta_f));
	mpcomplex res;
	mpreal rerr1, rerr2 = 0_mp;

	if( abs(x) <= 1_mp ) {
		// this is Eq. 10 of Karzas & Latter
		mpcomplex a( l+1, -eta_f );
		mpcomplex b( l+1, -eta_i );
		mpreal c( 2*(l+1) );
		res = F21( a, b, c, x, rerr1 );
		cg = 0_mp;
	}
	else {
		// here Eq. 15.3.4 of Abramowitz & Stegun
		// has been applied to Eq. 11 of Karzas & Latter
		mpcomplex a( l+1, -eta_f );
		mpcomplex b( -l, -eta_f );
		mpcomplex c( 1, eta_imf );
		mpcomplex r1 = F21( a, c-b, c, 1_mp/(1_mp - x), rerr1 );
		mpcomplex r2 = conj(r1);
		r1 *= pow(1_mp - x,mpcomplex(-l-1,eta_f));
		r2 *= pow(1_mp - x,mpcomplex(-l-1,eta_i));
		mpcomplex gflog = cg2 - conj(cgi) - cgf;
		mpreal fac = (abs(cg2.imag())+abs(cgf.imag())+abs(cgi.imag()))/abs(gflog.imag());
		// for eps_i around 1e-17 or lower, the real part of gflog will become so large
		// that the call to exp() will overflow, so we pass the real part separately to
		// Il() where it will nearly cancel the real part of the sum cgi + cgf.
		cg = gflog.real();
		gflog.real( 0_mp );
		mpcomplex gf = exp( gflog );
		r1 *= gf;
		r2 *= conj(gf);
		gf = ( l == 1 ) ? 6_mp : 1_mp;
		res = gf*(r1 + r2);
		rerr2 = fac*machine_epsilon(mpreal::default_prec)*max(abs(r1),abs(r2))/abs(r1+r2);
	}
	res *= pow(xp,mpcomplex(0,eta_i+eta_f));
	rerr = max(hypot(rerr1,rerr2),abs(res.imag()/res.real()));
	if( !finite(res.real()) || rerr > max_rerr )
		throw loss_of_precision();
	return res.real();
}

// F21: compute the hypergeometric function 2F1(a,b;c;z) for complex arguments
mpcomplex F21(const mpcomplex& a, const mpcomplex& b, const mpcomplex& c, const mpreal& x, mpreal& rerr)
{
	// x = -4*eta_i*eta_f/(eta-i-eta_f)^2 is a number such that -1 <= x < 0
	// for x close to -1, the evaluation of F21 becomes exceedingly slow
	// so here we transform z = x/(x-1), so that 0 < z <= 1/2
	// this is Eq. 15.3.4 of Abramowitz & Stegun
	if( x < 0_mp ) {
		mpcomplex res = F21( a, c-b, c, x/(x-1_mp), rerr );
		return pow( 1_mp - x, -a )*res;
	}

	mpcomplex val( 1 );
	mpcomplex powx( 1 );
	mpcomplex fac( 1 );
	mpreal m( 1 );
	mpreal maxv( 0 );
	mpreal absv, absv2;
	bool lgCrit[3];
	lgCrit[0] = lgCrit[1] = lgCrit[2] = false;
	while( true ) {
		lgCrit[0] = lgCrit[1];
		lgCrit[1] = lgCrit[2];
		powx *= x;
		fac *= (a+m-1_mp)*(b+m-1_mp)/((c+m-1_mp)*m);
		mpcomplex term = fac*powx;
		m += 1_mp;
		absv = abs(term);
		maxv = max(absv,maxv);
		absv2 = abs(val);
		if( absv2 > 1_mp/powi(machine_epsilon(mpreal::default_prec),3) ) {
			throw loss_of_precision();
		}
		lgCrit[2] = ( absv/absv2 < machine_epsilon(mpreal::default_prec) );
		rerr = sqrt(m)*maxv/absv2*machine_epsilon(mpreal::default_prec);
		// see p. 32 of Master's thesis of John Pearson (2009)
		if( lgCrit[0] && lgCrit[1] && lgCrit[2] )
			break;
		val += term;
	}
	if( rerr > max_rerr )
		throw loss_of_precision();
	return val;
}

struct spouge
{
	int a;
	vector<mpreal> ck;
};

class clgamma_coef
{
	map<mp_prec_t, const spouge*> p_coef;
	const spouge* p_init_spouge(mp_prec_t prec)
	{
		spouge* sp = new spouge;

		mpreal ln_twopi = log(4_mp*asin(1_mp));
		mpreal P = -log(machine_epsilon(prec))/ln_twopi;
		mpreal a = ceil(P - log(P)/(2_mp*ln_twopi) - 0.5_mp);
		sp->a = int(a);

		mp_prec_t save = mpreal::default_prec;
		// this algorithm suffers from cancellation errors
		// so we use double precision internally
		mpreal::set_default_prec(2*prec);

		mpreal e = exp(1_mp);

		// the Spouge formula, see
		// http://en.wikipedia.org/wiki/Spouge%27s_approximation

		sp->ck.resize(a);
		mpreal sgn( 1 );
		mpreal fac = exp(a-1_mp);
		for( int kk=1; kk < sp->a; ++kk ) {
			mpreal k( kk );
			sp->ck[kk] = sgn*powi(a-k,kk)/sqrt(a-k)*fac;
			sgn = -sgn;
			fac /= e*k;
		}
		mpreal::set_default_prec(save);
		return sp;
	}
	void p_init(mp_prec_t prec)
	{
		auto p = p_coef.find(prec);
		if( p == p_coef.end() )
			p_coef[prec] = p_init_spouge(prec);
	}
public:
	clgamma_coef() {}
	~clgamma_coef()
	{
		for( auto p = p_coef.begin(); p != p_coef.end(); ++p )
			delete (*p).second;
		p_coef.clear();
	}
	int a(mp_prec_t prec)
	{
		p_init(prec);
		return p_coef[prec]->a;
	}
	mpreal ck(mp_prec_t prec, size_t i)
	{
		return p_coef[prec]->ck[i];
	}
} cc;
	
// clgamma: compute the log of the complex gamma function
mpcomplex clgamma(const mpcomplex& z0)
{
	if( z0.real() <= 1_mp )
		return clgamma(z0+1_mp) - log(z0);

	mp_prec_t save = mpreal::default_prec;
	// this algorithm suffers from cancellation errors
	// so we use double precision internally
	mpreal::set_default_prec(2*save);

	// keep declaration and initialization separate in order
	// to assure that z indeed has double precision...
	mpcomplex z;
	z = z0 - 1_mp;

	// the Spouge formula, see
	// http://en.wikipedia.org/wiki/Spouge%27s_approximation
	mpreal twopi = 4_mp*asin(1_mp);
	mpcomplex sum( sqrt(twopi), 0 );
	int aa = cc.a(save);
	mpreal a( aa );
	mpreal maxv( 0 );
	for( int kk=1; kk < aa; ++kk ) {
		mpreal k( kk );
		sum += cc.ck(save,kk)/(z+k);
		maxv = max( maxv, abs(sum) );
	}
	mpcomplex gl = (z+0.5_mp)*log(z+a) - z - a + log(sum);
	assert( abs(sum)/maxv > machine_epsilon(save) );
	mpreal::set_default_prec(save);
	return gl;
}

mpreal eta_sub(const mpreal& eps_i, const mpreal& w)
{
	mpreal eta_i = 1_mp/sqrt(eps_i);
	mpreal x = w/eps_i;
	mpreal res;
	if( x < 0.1_mp ) {
		int i = 1;
		mpreal fac = x/2_mp;
		res = 0_mp;
		do {
			res += fac;
			++i;
			fac *= -mpreal(2*i-1)/mpreal(2*i)*x;
		} while( abs(fac/res) > machine_epsilon(mpreal::default_prec) );
		res *= eta_i;
	}
	else {
		res = eta_i - 1_mp/sqrt(eps_i+w);
	}
	return res;
}
